"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = y[op[0] & 2 ? "return" : op[0] ? "throw" : "next"]) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [0, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
var helpers_1 = require("../helpers");
var MySqlGrammar_1 = require("./grammars/MySqlGrammar");
var PostgresGrammar_1 = require("./grammars/PostgresGrammar");
var Expression_1 = require("./Expression");
var WhereType;
(function (WhereType) {
    WhereType[WhereType["Basic"] = 0] = "Basic";
    WhereType[WhereType["In"] = 1] = "In";
    WhereType[WhereType["Null"] = 2] = "Null";
    WhereType[WhereType["Nested"] = 3] = "Nested";
    WhereType[WhereType["Column"] = 4] = "Column";
    WhereType[WhereType["Raw"] = 5] = "Raw";
})(WhereType = exports.WhereType || (exports.WhereType = {}));
var Builder = /** @class */ (function () {
    function Builder(connection) {
        this._distinct = false;
        this._columns = ["*"];
        this._from = "";
        this._joins = [];
        this._wheres = [];
        this._groups = [];
        this._havings = [];
        this._orders = [];
        this._bindings = {
            Select: [],
            Join: [],
            Update: [],
            Where: [],
            Having: [],
            Values: [],
        };
        /**
         * All of the available clause operators
         */
        this.operators = [
            "=", "<", ">", "<=", ">=", "<>", "!=", "<=>",
            "like", "like binary", "not like", "between", "ilike",
            "&", "|", "^", "<<", ">>",
            "rlike", "regexp", "not regexp",
            "~", "~*", "!~", "!~*", "similar to",
            "not similar to", "not ilike", "~~*", "!~~*",
        ];
        this._lastSql = "";
        this.connection = connection;
    }
    Builder.prototype.newQuery = function () {
        return new Builder(this.connection);
    };
    Builder.prototype.select = function (columns) {
        if (columns === void 0) { columns = ["*"]; }
        this._columns = columns;
        return this;
    };
    Builder.prototype.addSelect = function (column) {
        this._columns.push(column);
        return this;
    };
    Builder.prototype.selectRaw = function (sql, bindings) {
        if (bindings === void 0) { bindings = []; }
        this.addBindingsSelect(bindings);
        return this.addSelect(new Expression_1.Expression(sql));
    };
    Builder.prototype.table = function (table) {
        this._from = table;
        return this;
    };
    Builder.prototype.join = function (table, first, operator, second, type) {
        if (operator === void 0) { operator = ""; }
        if (second === void 0) { second = ""; }
        if (type === void 0) { type = "inner"; }
        // If the first column is actually a function, we will assume the developer
        // wants to begin a nested on statement which is wrapped in parenthesis.
        if (typeof first === "function") {
            var query_1 = this.newQuery();
            query_1.table(this._from);
            query_1.whereNested(first);
            this.pushJoin(type, table, query_1);
            this.addBindingsJoin(query_1.bindings());
            return this;
        }
        if (3 === arguments.length) {
            second = operator;
            operator = "=";
        }
        if (undefined === operator || this.invalidOperator(operator)) {
            second = operator;
            operator = "=";
        }
        var query = this.newQuery();
        query.table(this._from);
        query.whereColumn(first, operator, second);
        this.pushJoin(type, table, query);
        return this;
    };
    Builder.prototype.crossJoin = function (table) {
        return this.pushJoin("cross", table);
    };
    Builder.prototype.leftJoin = function (table, first, operator, second) {
        if (operator === void 0) { operator = ""; }
        if (second === void 0) { second = ""; }
        return this.join(table, first, operator, second, "left");
    };
    Builder.prototype.rightJoin = function (table, first, operator, second) {
        if (operator === void 0) { operator = ""; }
        if (second === void 0) { second = ""; }
        return this.join(table, first, operator, second, "right");
    };
    Builder.prototype.where = function (column, operator, value, conj) {
        if (conj === void 0) { conj = "and"; }
        // If the columns is actually a function, we will assume the developer
        // wants to begin a nested where statement which is wrapped in parenthesis.
        if (typeof column === "function") {
            return this.whereNested(column, conj);
        }
        // Here we will make some assumptions about the operator. If only 2 values are
        // passed to the method, we will assume that the operator is an equals sign
        // and keep going. Otherwise, we'll require the operator to be passed in.
        if (2 === arguments.length) {
            value = operator;
            operator = "=";
        }
        // If the given operator is not found in the list of valid operators we will
        // assume that the developer is just short-cutting the '=' operators and
        // we will set the operators to '=' and set the values appropriately.
        if (typeof operator !== "string" || this.invalidOperator(operator)) {
            value = operator;
            operator = "=";
        }
        // If the value is "null", we will just assume the developer wants to add a
        // where null clause to the query. So, we will allow a short-cut here to
        // that method for convenience so the developer doesn't have to check.
        if (null === value) {
            if ("=" === operator) {
                return this.whereNull(column, conj);
            }
            return this.whereNotNull(column, conj);
        }
        this.pushWhereBasic(conj, column, operator, value);
        return this;
    };
    Builder.prototype.orWhere = function (column, operator, value) {
        return this.where(column, operator, value, "or");
    };
    Builder.prototype.whereColumn = function (first, operator, second, conj) {
        if (second === void 0) { second = ""; }
        if (conj === void 0) { conj = "and"; }
        // Here we will make some assumptions about the operator. If only 2 values are
        // passed to the method, we will assume that the operator is an equals sign
        // and keep going. Otherwise, we'll require the operator to be passed in.
        if (2 === arguments.length) {
            second = operator;
            operator = "=";
        }
        // If the given operator is not found in the list of valid operators we will
        // assume that the developer is just short-cutting the '=' operators and
        // we will set the operators to '=' and set the values appropriately.
        if (typeof operator !== "string" || this.invalidOperator(operator)) {
            second = operator;
            operator = "=";
        }
        this.pushWhereColumn(conj, first, operator, second);
        return this;
    };
    Builder.prototype.on = function (first, operator, second, conj) {
        if (second === void 0) { second = ""; }
        if (conj === void 0) { conj = "and"; }
        return this.whereColumn(first, operator, second, conj);
    };
    Builder.prototype.orOn = function (first, operator, second) {
        if (second === void 0) { second = ""; }
        return this.whereColumn(first, operator, second, "or");
    };
    Builder.prototype.orWhereColumn = function (first, operator, second, conj) {
        if (second === void 0) { second = ""; }
        if (conj === void 0) { conj = "and"; }
        return this.whereColumn(first, operator, second, "or");
    };
    Builder.prototype.whereBetween = function (column, values, conj, not) {
        if (conj === void 0) { conj = "and"; }
        if (not === void 0) { not = false; }
        return this;
    };
    Builder.prototype.whereRaw = function (sql, bindings, conj) {
        if (bindings === void 0) { bindings = []; }
        if (conj === void 0) { conj = "and"; }
        this.pushWhereRaw(conj, sql, bindings);
        return this;
    };
    Builder.prototype.orWhereRaw = function (sql, bindings) {
        if (bindings === void 0) { bindings = []; }
        return this.whereRaw(sql, bindings, "or");
    };
    Builder.prototype.whereNested = function (fn, conj) {
        if (conj === void 0) { conj = "and"; }
        var query = this.newQuery();
        query.table(this._from);
        fn(query);
        this.pushWhereNested(conj, query);
        return this;
    };
    Builder.prototype.orWhereNested = function (fn) {
        return this.whereNested(fn, "or");
    };
    Builder.prototype.whereNull = function (column, conj, not) {
        if (conj === void 0) { conj = "and"; }
        if (not === void 0) { not = false; }
        this.pushWhereNull(conj, column, not);
        return this;
    };
    Builder.prototype.orWhereNull = function (column, not) {
        if (not === void 0) { not = false; }
        return this.whereNull(column, "or", not);
    };
    Builder.prototype.whereNotNull = function (column, conj) {
        if (conj === void 0) { conj = "and"; }
        return this.whereNull(column, conj, true);
    };
    Builder.prototype.orWhereNotNull = function (column) {
        return this.whereNull(column, "or", true);
    };
    Builder.prototype.whereIn = function (column, values, conj, not) {
        if (conj === void 0) { conj = "and"; }
        if (not === void 0) { not = false; }
        this.pushWhereIn(conj, column, values, not);
        return this;
    };
    Builder.prototype.orWhereIn = function (column, values, not) {
        if (not === void 0) { not = false; }
        return this.whereIn(column, values, "or", not);
    };
    Builder.prototype.whereNotIn = function (column, values, conj) {
        if (conj === void 0) { conj = "and"; }
        return this.whereIn(column, values, conj, true);
    };
    Builder.prototype.orWhereNotIn = function (column, values) {
        return this.whereIn(column, values, "or", true);
    };
    Builder.prototype.groupBy = function (column) {
        this._groups.push({
            Column: column,
        });
        return this;
    };
    Builder.prototype.groupByRaw = function (sql) {
        this._groups.push({
            Sql: sql,
        });
        return this;
    };
    Builder.prototype.having = function (column, operator, value, conj) {
        if (conj === void 0) { conj = "and"; }
        // Here we will make some assumptions about the operator. If only 2 values are
        // passed to the method, we will assume that the operator is an equals sign
        // and keep going. Otherwise, we'll require the operator to be passed in.
        if (2 === arguments.length) {
            value = operator;
            operator = "=";
        }
        // If the given operator is not found in the list of valid operators we will
        // assume that the developer is just short-cutting the '=' operators and
        // we will set the operators to '=' and set the values appropriately.
        if (typeof operator !== "string" || this.invalidOperator(operator)) {
            value = operator;
            operator = "=";
        }
        this.pushHavingBasic(conj, column, operator, value);
        return this;
    };
    Builder.prototype.orHaving = function (column, operator, value) {
        return this.having(column, operator, value, "or");
    };
    Builder.prototype.havingRaw = function (sql, bindings, conj) {
        if (bindings === void 0) { bindings = []; }
        if (conj === void 0) { conj = "and"; }
        this.pushHavingRaw(conj, sql, bindings);
        return this;
    };
    Builder.prototype.orHavingRaw = function (sql, bindings) {
        if (bindings === void 0) { bindings = []; }
        return this.havingRaw(sql, bindings, "or");
    };
    Builder.prototype.orderBy = function (column, direction) {
        if (direction === void 0) { direction = "asc"; }
        this._orders.push({
            Column: column,
            Direction: direction.toLowerCase(),
        });
        return this;
    };
    Builder.prototype.orderByRaw = function (sql) {
        this._orders.push({
            Sql: sql,
        });
        return this;
    };
    Builder.prototype.limit = function (n) {
        this._limit = n;
        return this;
    };
    Builder.prototype.offset = function (n) {
        this._offset = n;
        return this;
    };
    Builder.prototype.bindings = function () {
        return this.mergeBindings();
    };
    Builder.prototype.toSql = function () {
        return this.grammar().compileSelect(this);
    };
    /**
     * Exec update clause
     * @param map
     */
    Builder.prototype.update = function (map) {
        return __awaiter(this, void 0, void 0, function () {
            var values, sql, bindings, res;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        values = helpers_1.objectHelper.values(map);
                        this.addBindingsUpdate(values);
                        sql = this.grammar().compileUpdate(this, map);
                        this._lastSql = sql;
                        bindings = this.bindings();
                        return [4 /*yield*/, this.connection.rawSql(sql, bindings)];
                    case 1:
                        res = _a.sent();
                        return [2 /*return*/, this.affectedRows(res)];
                }
            });
        });
    };
    Builder.prototype.insert = function (map) {
        return __awaiter(this, void 0, void 0, function () {
            var maps, values, bindings, sql, res;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        maps = [];
                        if (Array.isArray(map)) {
                            maps = map;
                        }
                        else {
                            maps = [map];
                        }
                        values = [];
                        maps.forEach(function (map) {
                            values = helpers_1.arrayHelper.merge(values, helpers_1.objectHelper.values(map));
                        });
                        bindings = this.clearBindings().addBindingsValues(values).bindings();
                        sql = this.grammar().compileInsert(this, maps);
                        this._lastSql = sql;
                        return [4 /*yield*/, this.connection.rawSql(sql, bindings)];
                    case 1:
                        res = _a.sent();
                        return [2 /*return*/, this.affectedRows(res)];
                }
            });
        });
    };
    Builder.prototype.insertGetId = function (map, sequence) {
        if (sequence === void 0) { sequence = "id"; }
        return __awaiter(this, void 0, void 0, function () {
            var values, bindings, sql, res;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        values = helpers_1.objectHelper.values(map);
                        bindings = this.clearBindings().addBindingsValues(values).bindings();
                        sql = this.grammar().compileInsertGetId(this, map, sequence);
                        this._lastSql = sql;
                        return [4 /*yield*/, this.connection.rawSql(sql, bindings)];
                    case 1:
                        res = _a.sent();
                        switch (this.connection.driverType()) {
                            case 0 /* mysql */: {
                                return [2 /*return*/, res.insertId];
                            }
                            case 1 /* pgsql */: {
                                return [2 /*return*/, res.rows[0][sequence]];
                            }
                        }
                        return [2 /*return*/, 0];
                }
            });
        });
    };
    Builder.prototype.delete = function () {
        return __awaiter(this, void 0, void 0, function () {
            var sql, bindings, res;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        sql = this.grammar().compileDelete(this);
                        this._lastSql = sql;
                        bindings = this.bindings();
                        return [4 /*yield*/, this.connection.rawSql(sql, bindings)];
                    case 1:
                        res = _a.sent();
                        return [2 /*return*/, this.affectedRows(res)];
                }
            });
        });
    };
    Builder.prototype.truncate = function () {
        return __awaiter(this, void 0, void 0, function () {
            var sql, res;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        sql = this.grammar().compileTruncate(this);
                        this._lastSql = sql;
                        return [4 /*yield*/, this.connection.rawSql(sql)];
                    case 1:
                        res = _a.sent();
                        return [2 /*return*/, this.affectedRows(res)];
                }
            });
        });
    };
    /**
     * Query select clause and fetch the first one
     * @param columns
     */
    Builder.prototype.first = function (columns) {
        if (columns === void 0) { columns = ["*"]; }
        return __awaiter(this, void 0, void 0, function () {
            var original, records;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        original = this._columns;
                        if (undefined === original || helpers_1.arrayHelper.equal(original, ["*"])) {
                            this._columns = columns;
                        }
                        this.limit(1);
                        return [4 /*yield*/, this.runSelect()];
                    case 1:
                        records = _a.sent();
                        if (0 === records.length) {
                            return [2 /*return*/, null];
                        }
                        return [2 /*return*/, records[0]];
                }
            });
        });
    };
    /**
     * Query select clause and fetch records
     * @param columns
     */
    Builder.prototype.get = function (columns) {
        if (columns === void 0) { columns = ["*"]; }
        var original = this._columns;
        if (undefined === original || helpers_1.arrayHelper.equal(original, ["*"])) {
            this._columns = columns;
        }
        return this.runSelect();
    };
    /**
     * Fetch the column's value of first record
     * @param column
     */
    Builder.prototype.value = function (column) {
        return __awaiter(this, void 0, void 0, function () {
            var record, firstKey;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.first([column])];
                    case 1:
                        record = _a.sent();
                        if (null === record) {
                            return [2 /*return*/, null];
                        }
                        firstKey = Object.keys(record)[0];
                        return [2 /*return*/, record[firstKey]];
                }
            });
        });
    };
    /**
     * Fetch the column's values of records
     * @param column
     * @param key
     */
    Builder.prototype.pluck = function (column, key) {
        if (key === void 0) { key = ""; }
        return __awaiter(this, void 0, void 0, function () {
            var pluck_1, records_1, map, records;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!(0 === key.length)) return [3 /*break*/, 2];
                        pluck_1 = [];
                        return [4 /*yield*/, this.get([column])];
                    case 1:
                        records_1 = _a.sent();
                        records_1.forEach(function (record) {
                            pluck_1.push(record[column]);
                        });
                        return [2 /*return*/, pluck_1];
                    case 2:
                        map = {};
                        return [4 /*yield*/, this.get([column, key])];
                    case 3:
                        records = _a.sent();
                        records.forEach(function (record) {
                            map[record[key]] = record[column];
                        });
                        return [2 /*return*/, map];
                }
            });
        });
    };
    Builder.prototype.lastSql = function () {
        return this._lastSql;
    };
    Builder.prototype.count = function (column) {
        if (column === void 0) { column = "*"; }
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.aggregate("count", column)];
                    case 1: return [2 /*return*/, _a.sent()];
                }
            });
        });
    };
    Builder.prototype.sum = function (column) {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.aggregate("sum", column)];
                    case 1: return [2 /*return*/, _a.sent()];
                }
            });
        });
    };
    Builder.prototype.max = function (column) {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.aggregate("max", column)];
                    case 1: return [2 /*return*/, _a.sent()];
                }
            });
        });
    };
    Builder.prototype.min = function (column) {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.aggregate("min", column)];
                    case 1: return [2 /*return*/, _a.sent()];
                }
            });
        });
    };
    Builder.prototype.avg = function (column) {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.aggregate("avg", column)];
                    case 1: return [2 /*return*/, _a.sent()];
                }
            });
        });
    };
    Builder.prototype.affectedRows = function (res) {
        switch (this.connection.driverType()) {
            case 0 /* mysql */: {
                return res.affectedRows;
            }
            case 1 /* pgsql */: {
                return Number(res.rowCount);
            }
        }
        return 0;
    };
    Builder.prototype.aggregate = function (func, column) {
        return __awaiter(this, void 0, void 0, function () {
            var query, record, firstKey;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        query = Object.create(this);
                        return [4 /*yield*/, query.select().setAggregate(func, column).runSelect()];
                    case 1:
                        record = (_a.sent())[0];
                        this._lastSql = query.lastSql();
                        console.log(this._lastSql);
                        firstKey = Object.keys(record)[0];
                        return [2 /*return*/, Number(record[firstKey])];
                }
            });
        });
    };
    Builder.prototype.setAggregate = function (func, column) {
        this._aggregate = {
            Func: func,
            Column: column,
        };
        return this;
    };
    Builder.prototype.grammar = function () {
        if (undefined === this._grammar) {
            switch (this.connection.driverType()) {
                case 0 /* mysql */: {
                    this._grammar = new MySqlGrammar_1.MySqlGrammar();
                    break;
                }
                case 1 /* pgsql */: {
                    this._grammar = new PostgresGrammar_1.PostgresGrammar();
                    break;
                }
            }
        }
        return this._grammar;
    };
    Builder.prototype.runSelect = function () {
        return __awaiter(this, void 0, void 0, function () {
            var sql, bindings, records;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        sql = this.toSql();
                        this._lastSql = sql;
                        bindings = this.bindings();
                        return [4 /*yield*/, this.connection.rawSql(sql, bindings)];
                    case 1:
                        records = _a.sent();
                        return [2 /*return*/, this.toRecords(records)];
                }
            });
        });
    };
    Builder.prototype.toRecords = function (res) {
        switch (this.connection.driverType()) {
            case 0 /* mysql */: {
                return res;
            }
            case 1 /* pgsql */: {
                return res.rows;
            }
        }
        return {};
    };
    Builder.prototype.pushJoin = function (type, table, query) {
        this._joins.push({
            Type: type,
            Table: table,
            Query: query,
        });
        return this;
    };
    Builder.prototype.pushHavingBasic = function (conj, column, operator, value) {
        this._havings.push({
            Conjunction: conj,
            Column: column,
            Operator: operator,
            Value: value,
        });
        this.addBindingsHaving(value);
        return this;
    };
    Builder.prototype.pushHavingRaw = function (conj, sql, bindings) {
        if (bindings === void 0) { bindings = []; }
        this._havings.push({
            Conjunction: conj,
            Sql: sql,
        });
        this.addBindingsHaving(bindings);
        return this;
    };
    Builder.prototype.pushWhereBasic = function (conj, column, operator, value) {
        this._wheres.push({
            Type: WhereType.Basic,
            Conjunction: conj.toLowerCase(),
            Column: column,
            Operator: operator,
            Value: value,
        });
        this.addBindingsWhere(value);
        return this;
    };
    Builder.prototype.pushWhereColumn = function (conj, first, operator, second) {
        this._wheres.push({
            Type: WhereType.Column,
            Conjunction: conj,
            Column: first,
            Operator: operator,
            Value: second,
        });
        return this;
    };
    Builder.prototype.pushWhereIn = function (conj, column, values, not) {
        this._wheres.push({
            Type: WhereType.In,
            Conjunction: conj.toLowerCase(),
            Column: column,
            Values: values,
            Not: not,
        });
        this.addBindingsWhere(values);
        return this;
    };
    Builder.prototype.pushWhereNull = function (conj, column, not) {
        this._wheres.push({
            Type: WhereType.Null,
            Conjunction: conj.toLowerCase(),
            Column: column,
            Not: not,
        });
        return this;
    };
    Builder.prototype.pushWhereNested = function (conj, query) {
        this._wheres.push({
            Type: WhereType.Nested,
            Conjunction: conj.toLowerCase(),
            Query: query,
        });
        this.addBindingsWhere(query._bindings.Where);
        return this;
    };
    Builder.prototype.pushWhereRaw = function (conj, sql, bindings) {
        if (bindings === void 0) { bindings = []; }
        this._wheres.push({
            Type: WhereType.Raw,
            Conjunction: conj.toLowerCase(),
            Sql: sql,
        });
        this.addBindingsWhere(bindings);
        return this;
    };
    Builder.prototype.addBindingsHaving = function (value) {
        var _this = this;
        if (Array.isArray(value)) {
            value.forEach(function (v) {
                _this.addBindingsHaving(v);
            });
            return this;
        }
        if (value instanceof Expression_1.Expression === false) {
            this._bindings.Having.push(value);
        }
        return this;
    };
    Builder.prototype.addBindingsUpdate = function (values) {
        var _this = this;
        values.forEach(function (v) {
            if (v instanceof Expression_1.Expression === false) {
                _this._bindings.Update.push(v);
            }
        });
        return this;
    };
    /**
     * insert values
     * @param values
     */
    Builder.prototype.addBindingsValues = function (values) {
        var _this = this;
        values.forEach(function (v) {
            if (v instanceof Expression_1.Expression === false) {
                _this._bindings.Values.push(v);
            }
        });
        return this;
    };
    Builder.prototype.addBindingsWhere = function (value) {
        var _this = this;
        if (Array.isArray(value)) {
            value.forEach(function (v) {
                _this.addBindingsWhere(v);
            });
            return this;
        }
        if (value instanceof Expression_1.Expression === false) {
            this._bindings.Where.push(value);
        }
        return this;
    };
    Builder.prototype.addBindingsJoin = function (value) {
        var _this = this;
        if (Array.isArray(value)) {
            value.forEach(function (v) {
                _this.addBindingsJoin(v);
            });
            return this;
        }
        if (value instanceof Expression_1.Expression === false) {
            this._bindings.Join.push(value);
        }
        return this;
    };
    Builder.prototype.addBindingsSelect = function (values) {
        var _this = this;
        values.forEach(function (v) {
            if (v instanceof Expression_1.Expression === false) {
                _this._bindings.Select.push(v);
            }
        });
        return this;
    };
    Builder.prototype.mergeBindings = function () {
        return helpers_1.arrayHelper.merge(this._bindings.Select, this._bindings.Join, this._bindings.Update, this._bindings.Where, this._bindings.Having, this._bindings.Values);
    };
    Builder.prototype.clearBindings = function () {
        this._bindings = {
            Select: [],
            Join: [],
            Update: [],
            Where: [],
            Having: [],
            Values: [],
        };
        return this;
    };
    Builder.prototype.invalidOperator = function (operator) {
        operator = operator.toLowerCase();
        return -1 === this.operators.indexOf(operator) && -1 === this.grammar().getOperators().indexOf(operator);
    };
    return Builder;
}());
exports.Builder = Builder;
//# sourceMappingURL=Builder.js.map