import { Builder, Order, Where, Group, Having, Map } from "../Builder";
import { Expression } from "../Expression";
export declare abstract class Grammar {
    protected operators: string[];
    getOperators(): string[];
    /**
     * Build the select statement sql
     * @param query
     */
    compileSelect(query: Builder): string;
    compileUpdate(query: Builder, map: Map): string;
    compileInsert(query: Builder, maps: Map[]): string;
    compileInsertGetId(query: Builder, map: Map, sequence: string): string;
    compileDelete(query: Builder): string;
    compileTruncate(query: Builder): string;
    /**
     * Convert an array of column names into a delimited string
     * @param columns
     */
    columnize(columns: [string | Expression]): string;
    /**
     * Wrap a value in keyword identifiers
     * @param value
     */
    wrap(value: string | Expression): string;
    /**
     * Wrap a value that has an alias
     * @param value
     */
    wrapAliasedValue(value: string): string;
    /**
     * Wrap a single string in keyword identifiers
     * @param value
     */
    wrapSingleValue(value: string): string;
    /**
     * Get the appropriate query parameter place-holder
     */
    parameter(value?: any): string;
    /**
     * Create query parameter place-holders for an array
     * @param params
     */
    parameterize(params: any[]): string;
    /**
     * Compile an aggregated select clause
     * @param query
     */
    protected compileAggregate(query: Builder): string;
    /**
     * Compile the "select *" portion of the query
     * @param query
     */
    protected compileColumn(query: Builder): string;
    /**
     * Compile the "from" portion of the query
     * @param query
     */
    protected compileFrom(query: Builder): string;
    /**
     * Compile the "join" portions of the query
     * @param query
     */
    protected compileJoins(query: Builder): string;
    protected compileJoinsToArray(query: Builder): string[];
    /**
     * Compile the "where" portions of the query
     * @param query
     */
    protected compileWheres(query: Builder): string;
    /**
     * Get an array of all the where clauses for the query
     * @param query
     */
    protected compileWheresToArray(query: Builder): string[];
    /**
     * Format the where clause statements into one string
     * @param wheresArr
     */
    protected concatenateWhereClauses(wheresArr: string[]): string;
    protected whereBasic(where: Where): string;
    protected whereColumn(where: Where): string;
    protected whereRaw(where: Where): string;
    protected whereNull(where: Where): string;
    protected whereNotNull(where: Where): string;
    protected whereIn(where: Where): string;
    protected whereNotIn(where: Where): string;
    protected whereNested(where: Where): string;
    /**
     * Compile the "group by" portions of the query
     * @param query
     */
    protected compileGroups(query: Builder): string;
    /**
     * Compile the query groups to an array
     * @param groups
     */
    protected compileGroupsToArray(groups: Group[]): string[];
    /**
     * Compile the "having" portions of the query
     * @param query
     */
    protected compileHavings(query: Builder): string;
    protected compileHavingsToArray(havings: Having[]): string[];
    protected concatenateHavingClause(havingsArr: string[]): string;
    /**
     * Compile the "order by" portions of the query
     * @param query
     */
    protected compileOrders(query: Builder): string;
    /**
     * Compile the query orders to an array
     * @param orders
     */
    protected compileOrdersToArray(orders: Order[]): string[];
    /**
     * Compile the "limit" portions of the query
     * @param query
     */
    protected compileLimit(query: Builder): string;
    /**
     * Compile the "offset" portions of the query
     * @param query
     */
    protected compileOffset(query: Builder): string;
}
